/**
 * The abstract property class.
 * <p>
 * The generic property object.  This object has methods to handle the property stacks,
 * used when traversing the graphs.  The stacks are class variables.
 * <p>
 * We encode the property "type" in the property as an integer, which tells us which stack to use.  There is 
 * probably a better way to do this, but this is pretty straightforward. Might be a cool way 
 * to do it using reflection, but I'm not going to worry about it here.
 * <p>
 * <b>NOTE:</b>  In theory, we would like to refresh the window when a property changes.  However, do not
 * do that in this assignment.  Properties (especially transformations) turn out to be useful for 
 * computing intermediate values, and causing a refresh for every modification would be problematic.
 * Even worse, if the property is updated in a dynamic property function (which would be common), 
 * the display function would overload with updates.
 *<p>
 * Therefore, if you change a property, either call repaint(), or reassign the property.
 *
 * @version 	1.00
 * @author      Blair MacIntyre
 */
import java.util.*;

public abstract class jooglProperty extends jooglObject {
    /** The id of the Transformation property */
    protected static final int PROP_TRANSFORM = 0;
    /** The id of the Material property */
    protected static final int PROP_MATERIAL = 1;
    /** The id of the DrawMode property */
    protected static final int PROP_DRAWMODE = 2;
    /** The id of the ShadeMode property */
    protected static final int PROP_SHADEMODE = 3;
    /** The id of the Pickable property */
    protected static final int PROP_PICKABLE = 4;

    /** The number of properties */
    protected static final int NUM_PROPERTIES = 5;
    
    /** The static stacks for property inheritance */
    protected static Stack[] props = new Stack[NUM_PROPERTIES];
     
    /** Have the static stacks been initialized? */
    protected static boolean init = false;

    /** What type of property am I?  Will be set when a property is created */
    protected int type = -1;
        
    /** 
     * The class constructor.
     * <p>
     * Initialize all of the internal class data, creating the inheritance stacks.
     *
     * @param the type of the property
     */     
    public jooglProperty(int _type)
    {
        super();
        type = _type;
        if (!init) 
        {
            for (int i=0; i<NUM_PROPERTIES; i++)
                props[i] = new Stack();
            
            init = true;
        }
    }

    /**
     * Push this object onto the property stack of the appropriate type, applying the new value.
     */
    protected void push()
    {
        props[type].push(this);
        // update the state for the new property
        apply();
    }
    
    /**
     * Pop the property stack of this property type, applying the old value.
     */
    protected void pop()
    {
        props[type].pop();
        // restore the previous state
        
        if (!empty())
            peek().apply();
    }

    /**
     * Is the stack for this property empty?
     * @return true if the stack is empty
     */
    protected boolean empty()
    {
        return props[type].empty();
    }

    /**
     * Is the stack for the specified property type empty?
     *
     * @param t the property type id
     * @return true if the stack is empty
     */
    protected static boolean empty(int t)
    {
        return props[t].empty();
    }
    
    /**
     * Get the top of the stack for the property of this object type.
     * @return the property on the top of the stack
     */
    protected jooglProperty peek()
    {
        if (empty())
            return null;
            
        return (jooglProperty)(props[type].peek());
    }
    
    /**
     * Get the top of the stack for the property of the specified type.
     * @return the property on the top of the stack
     */
    protected static jooglProperty peek(int t)
    {
        if (empty(t))
            return null;
            
        return (jooglProperty)(props[t].peek());
    }
    
    /** 
     * Is this property object dynamic or static?
     * <p>
     * By default, properties are static.  This will be overridden in the dynamic subclasses. 
     *
     * @return true if the property is dynamic.
     */
    protected boolean isDynamic()
    {	
        return false;
    }
    
    /** Internal Method, to set the OpenGL state for this property.  Called when the stacks are pushed and popped.
     * <p>
     * Only put things here that can be called multiple times safely.
     */
    protected abstract void apply();
}
