import java.util.*;
import gl4java.*;

/**
 * The transformation property node.
 * <p>
 * Transformation properties can either be constant or time-variant
 * functions, as with all the properties.  Unlike the material,
 * shademodel and drawmode, the transformation properties COMPOSE as the 
 * graph is traversed: at any given node, the "current transformation"
 * is the composition of all transformations on the path from the root
 * node to the node in question.
 * <p>
 * If a node does not have a transformation property, it is equivalent to the
 * identity matrix. 
 * <p>
 * We provide a set of functions to compose common
 * transformations (translate, scale, rotate) with the current value. 
 * We also provide functions to replace or compose arbitrary transformation
 * matrices with the current transformation property.  
 * <p>
 * setIdentity() reinitializes the current transformation
 * property to the identity matrix.
 * <p>
 * If a node has a time-variant function for its transformation
 * property, the functions to change the transformation have no effect.
 * <p>
 * @version 	1.00
 * @author      Blair MacIntyre
 */
public class jooglTransform extends jooglProperty {
    /** The current transformation matrix value */
    protected double[] matrix = new double[16];
    
    /** A useful value, the identity matrix */
    protected static final double[] identity = {1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0,
                                                0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0};

    /** 
     * Create a transformation containing the identity matrix.
     */
    public jooglTransform()
    {
        super(PROP_TRANSFORM);
        setIdentity();
    }
    
    /** 
     * Create a transformation identical to the supplied Transform.
     *
     * @param t a jooglTransform that this one should copy.
     */
    public jooglTransform(jooglTransform t)
    {
        super(PROP_TRANSFORM);
        setTransform(t.matrix);
    }
    
    /** 
     * Create a transformation identical to the supplied matrix.
     *
     * @param t a matrix that this one should copy.
     */
    public jooglTransform(double[] m)
    {
        super(PROP_TRANSFORM);
        setTransform(m);
    }
    
    /** 
     * Set the transformation matrix to be identical to the supplied matrix.
     *
     * @param t a matrix that this one should copy.
     * @return a reference to itself (this) so transformation commands can be chained.
     */
    public jooglTransform setTransform(double[] m)
    {
	for (int i=0; i<16; i++)
            matrix[i] = m[i];
        //redrawWindows();
        return this;
    }
    
    /** 
     * Set the transformation matrix to be identical to the supplied transformation.
     *
     * @param t a transform that this one should copy.
     * @return a reference to itself (this) so transformation commands can be chained.
     */
    public jooglTransform setTransform(jooglTransform m)
    {
        setTransform(m.matrix);
        return this;
    }

    /** 
     * Get the transformation matrix for this transform.
     *
     * @return a 16 element array containing the transformation
     */
    public double [] getTransform()
    {
        double[] m = new double[16];
	for (int i=0; i<16; i++)
            matrix[i] = m[i];
        return m;
    }
        
    /** 
     * Compose a translation onto the current transformation matrix.
     *
     * @param tx the x component of the translation.
     * @param ty the y component of the translation.
     * @param tz the z component of the translation.
     * @return a reference to itself (this) so transformation commands can be chained.
     */
    public jooglTransform translate(double tx, double ty, double tz)
    {
        /* if it's a null operation, don't bother! */
        if ((tx == 0.0) && (ty == 0.0) && (tz == 0.0))
            return this;

        /* do a specialized matrix multiply with current value */
        for(int i = 0; i < 4; ++i ) 
            matrix[12+i] += tx * matrix[i] + ty * matrix[4+i] + tz * matrix[8+i];

        //redrawWindows();
        return this;
    }

    /** 
     * Compose a scale onto the current transformation matrix.
     *
     * @param sx the x component of the scale.
     * @param sy the y component of the scale.
     * @param sz the z component of the scale.
     * @return a reference to itself (this) so transformation commands can be chained.
     */
    public jooglTransform scale(double sx, double sy, double sz)
    {
        /* if it's a null operation, don't bother! */
        if ((sx == 1.0) && (sy == 1.0) && (sz == 1.0))
            return this;

        /* do a specialized matrix multiply on TopOfMStack */
        for(int i = 0; i < 4; ++i ) 
        {
            matrix[i] *= sx;
            matrix[4+i] *= sy;
            matrix[8+i] *= sz;
        }

        //redrawWindows();
        return this;
    }

    /** 
     * Compose a rotation about X onto the end of the object transformation.
     *
     * @param angle the angle to rotate by, in degrees.
     * @return a reference to itself (this) so transformation commands can be chained.
     */
    public jooglTransform rotateX(double angle) 
    {
        /* if it's a null operation, don't bother doing it */
        if ((angle % 360.0) == 0.0)
            return this;

        double radians = (angle)*Math.PI/180.0;
        double c = Math.cos( radians );
        double s = Math.sin( radians );
        int r1 = 1;
        int r2 = 2;

        double temp;
        for(int i=0; i<4; i++) {
            temp = c*matrix[i+r1*4] + s*matrix[i+r2*4];
            matrix[i+r2*4] = c*matrix[i+r2*4] - s*matrix[i+r1*4];
            matrix[i+r1*4] = temp;
        }
        //redrawWindows();
        return this;
    }
    
    /** 
     * Compose a rotation about Y onto the end of the object transformation.
     *
     * @param angle the angle to rotate by, in degrees.
     * @return a reference to itself (this) so transformation commands can be chained.
     */
    public jooglTransform rotateY(double angle)
    {
        /* if it's a null operation, don't bother doing it */
        if ((angle % 360.0) == 0.0)
            return this;

        double radians = (angle)*Math.PI/180.0;
        double c = Math.cos( radians );
        double s = Math.sin( radians );
        int r1 = 2;
        int r2 = 0;

        double temp;
        for(int i=0; i<4; i++) {
            temp = c*matrix[i+r1*4] + s*matrix[i+r2*4];
            matrix[i+r2*4] = c*matrix[i+r2*4] - s*matrix[i+r1*4];
            matrix[i+r1*4] = temp;
        }
        //redrawWindows();
        return this;
    }
    
    /** 
     * Compose a rotation about Z onto the end of the object transformation.
     *
     * @param angle the angle to rotate by, in degrees.
     * @return a reference to itself (this) so transformation commands can be chained.
     */
    public jooglTransform rotateZ(double angle)
    {
        /* if it's a null operation, don't bother doing it */
        if ((angle % 360.0) == 0.0)
            return this;

        double radians = (angle)*Math.PI/180.0;
        double c = Math.cos( radians );
        double s = Math.sin( radians );
        int r1 = 0;
        int r2 = 1;

        double temp;
        for(int i=0; i<4; i++) {
            temp = c*matrix[i+r1*4] + s*matrix[i+r2*4];
            matrix[i+r2*4] = c*matrix[i+r2*4] - s*matrix[i+r1*4];
            matrix[i+r1*4] = temp;
        }
        //redrawWindows();
        return this;
    }

    /** 
     * Set the transformation to the identity matrix.
     *
     * @return a reference to itself (this) so transformation commands can be chained.
     */
    public jooglTransform setIdentity()
    {
    	for (int i=0; i<16; i++)
            matrix[i] = identity[i];
        //redrawWindows();
        return this;
    }
    
    /** 
     * Compose the supplied transformation matrix onto the end of the object transformation.
     *
     * @param m the transformation matrix.
     * @return a reference to itself (this) so transformation commands can be chained.
     */
    public jooglTransform multiply(double[] m)
    {
        double[] t = new double[16];
        for (int i = 0; i<4; i++)
           for (int j = 0; j<4; j++)
              t[(i*4)+j] = matrix[i*4]*m[j] + matrix[i*4+1]*m[4+j] + 
                           matrix[i*4+2]*m[8+j] + matrix[i*4+3]*m[12+j];
        for (int i = 0; i<16; i++)
            matrix[i] = t[i];
            
        //redrawWindows();
        return this;
    }
    
    /** 
     * Compose the supplied transformation onto the end of the object transformation.
     *
     * @param m the transformation.
     * @return a reference to itself (this) so transformation commands can be chained.
     */
    public jooglTransform multiply(jooglTransform m)
    {
        return multiply(m.matrix);
    }
    
    /** 
     * Get the value of the property at time t.
     *
     * @param time the time in milliseconds at which the property should be evaluated
     * @return the transformation in a 16 element array
     */
    public double[] getValue(long time)
    {
        double[] m = new double[16];
        
        // this is a constant property.  Override this method to create a time varying property.
    	for (int i=0; i<16; i++)
            m[i] = matrix[i];
         
       return m;
    }
    
    /**
     * Print the transformation matrix.
     * <p>
     * This method is useful for debugging.
     *
     * @return a reference to itself (this) so transformation commands can be chained.
     */
    public jooglTransform printMatrix()
    {
        System.out.println("[" + matrix[0] + ", " + matrix[1] + ", " + matrix[2] + ", " + matrix[3] + "]");
        System.out.println("[" + matrix[4] + ", " + matrix[5] + ", " + matrix[6] + ", " + matrix[7] + "]");
        System.out.println("[" + matrix[8] + ", " + matrix[9] + ", " + matrix[10] + ", " + matrix[11] + "]");
        System.out.println("[" + matrix[12] + ", " + matrix[13] + ", " + matrix[14] + ", " + matrix[15] + "]");
        return this;
    }
    
    /**
     * Internal Method, to cache the value of the property at time t in the property state variable.
     *
     * @param time the time in milliseconds at which the property should be evaluated
     */
    protected void cacheValue(long time)
    {
        // constrant, do nothing
    }
    
    /** Push a transformation onto the transformation stack.
     * <p>
     * Hint:  you need to override this method here.
     */
    protected void push()
    {
        gl.glPushMatrix();
        // must do this here, because we cannot do it over and over again
        gl.glMultMatrixd(matrix);
        super.push();
    }
    
    /** Pop a transformation off the transformation stack.
     * <p>
     * Hint:  you need to override this method here.
     */
    protected void pop()
    {
        super.pop();
        gl.glPopMatrix();
    }

    /** Internal Method, set the OpenGL state for this property.  Called when the stacks are pushed and popped.
     * <p>
     * Only put things here that can be called multiple times safely. 
     */
    protected void apply()
    {
    }
    
    /** Multiply a vector by the transformation.
     *
     * @param x the x component of the vector.
     * @param y the y component of the vector.
     * @param z the z component of the vector.
     * @param r a 3 element matrix that the transformed vector will be returned in.
     */
    protected void multVector(double x, double y, double z, double[] r)
    {
        r[0] = (x * matrix[0] + y * matrix[4]+ z * matrix[8]);
        r[1] = (x * matrix[1] + y * matrix[5]+ z * matrix[9]);
        r[2] = (x * matrix[2] + y * matrix[6]+ z * matrix[10]);
    }

    /** Multiply a point by the transformation.
     *
     * @param x the x component of the point.
     * @param y the y component of the point.
     * @param z the z component of the point.
     * @param r a 3 element matrix that the transformed point will be returned in.
     */
    protected void multPoint(double x, double y, double z, double[] r)
    {
        double w = x * matrix[3] + y * matrix[7]+ z * matrix[11] + matrix[15];
        r[0] = (x * matrix[0] + y * matrix[4]+ z * matrix[8] + matrix[12]) / w;
        r[1] = (x * matrix[1] + y * matrix[5]+ z * matrix[9] + matrix[13]) / w;
        r[2] = (x * matrix[2] + y * matrix[6]+ z * matrix[10] + matrix[14]) / w;
    }
}
